import torch
import numpy as np
import torch.nn.functional as F
from typing import Union, List, Tuple, Dict, Iterable
from .layers import PoolingStage2d, Linear, BatchNorm2d, Conv2d
from .layers import UNet2d
from .conversions import floats_to_tensor, tensor_to_index
from .conversions import segmented_to_tensor, images_to_tensor
from .nn_decorators import nn_classifier


@nn_classifier
class SemanticSegmenter(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        pass

    def __init__(self, in_channels: Tuple[int, int, int],
                 colors: Dict[str, Union[int, List[int]]],
                 downsampling: List[Union[dict, List[dict]]],
                 pooling: List[Tuple[int, int]],
                 upsampling: List[Union[dict, List[dict]]],
                 dense: List[dict] = [],
                 pooling_type: str = "Max",
                 padded: bool = True,
                 activation: str = "relu"):
        super().__init__()
        assert len(downsampling) == len(pooling) == len(upsampling)
        self.in_channels = in_channels
        self.categories = [c for c in colors.keys()]
        self.colors = [colors[c] for c in self.categories]
        self.input_norm = BatchNorm2d(in_channels)
        self.UNet = UNet2d(in_channels, downsampling, pooling, upsampling,
                           padded=padded)
        self.dense = PoolingStage2d(self.UNet.out_channels, dense,
                                    pooling_type=None, padded=True)
        self.output = Conv2d(self.dense.out_channels,
                             len(self.categories), (1, 1))

    def forward(self, X: torch.Tensor):
        X = self.input_norm(X)
        X = self.UNet(X)
        X = self.dense(X)
        return self.output(X)

    def data_to_tensor(self, X: Iterable[np.ndarray],
                       Y: Union[None, List[str]],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        x = images_to_tensor(X, self.device)
        y = None if Y is None else segmented_to_tensor(Y, self.colors,
                                                       self.device)
        w = None if weights is None else floats_to_tensor(weights)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        indexes = tensor_to_index(tensor)
        return np.array(self.colors)[indexes]

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[torch.Tensor, None]):
        if weights is None:
            return F.cross_entropy(y_pred, y_target)
        else:
            return F.nll_loss(F.log_softmax(y_pred) * weights, y_target)

    @property
    def dump(self):
        return {"type": type(self).__name__,
                "shapes": self.shapes,
                "shape in": self.shape_in,
                "categories": self.categories,
                "colors": self.colors,
                "input norm": self.input_norm.dump,
                "downsample": [c.dump for c in self.downsample],
                "upsample": [d.dump for d in self.upsample],
                "output": self.output.dump}



import torch
from .linear import Linear
from .activation import Activation
from .batch_norm import BatchNorm1d


class LinearStage(torch.nn.Module):
    """
    An 'LinearStage' module is a 'Linear' layer followed by an
    'Activation' layer and finally a 'BatchNorm1d' layer.

    Attributes
    ----------
    linear : Linear
        The linear layer
    activation : Activation
        The activation layer
    batch_norm : BatchNorm1d
        The batch normalization layer
    """

    @classmethod
    def from_dump(cls, dump: dict) -> 'LinearStage':
        """returns a 'LinearStage' obStageject from a dump"""
        assert dump["type"] == cls.__name__
        obj = cls(1, 1)
        obj.linear = Linear.from_dump(dump["linear"])
        obj.activation = Activation.from_dump(dump["activation"])
        obj.batch_norm = BatchNorm1d.from_dump(dump["batch norm"])
        return obj

    def __init__(self, in_features: int,
                 out_features: int = 8,
                 activation: str = "relu"):
        """
        Parameters
        ----------
        in_features : int
            The nuber of features in the input
        out_features : int
            The number of features in the ouput
        activation : str
            The name of the activation layer
        """
        super().__init__()
        self.linear = Linear(in_features, out_features)
        self.activation = Activation(activation)
        self.batch_norm = BatchNorm1d(out_features)

    def forward(self, input: torch.Tensor) -> torch.Tensor:
        return self.batch_norm(self.activation(self.linear(input)))

    @property
    def dump(self) -> dict:
        return {"type": type(self).__name__,
                "linear": self.linear.dump,
                "activation": self.activation.dump,
                "batch norm": self.batch_norm.dump}

    @property
    def in_features(self):
        return self.linear.in_features

    @property
    def out_features(self):
        return self.linear.out_features



import pandas as _pd
import numpy as _np
from ._decision_tree import RegressionTree as _RegressionTree
from machine_learning._templates import RegressorTemplate as _RegressorTemplate
from machine_learning._templates import ClassifierTemplate as _ClassifierTemplate
import machine_learning._utilities as _utils


class GradientBoostedRegressionTree(_RegressorTemplate):
    """Gradient Boosted Regression Tree"""

    def __init__(self):
        self.shrinkage = 0.1
        self.trees = []

    def _MSE(self, predicted: _np.ndarray, target: _np.ndarray) -> float:
        """MSE loss function"""
        return _np.mean((predicted-target)**2)

    def _grad(self, predicted: _np.ndarray,
              target: _np.ndarray) -> _np.ndarray:
        """Gradient of the MSE loss function:
        dLoss/df = d/df(sum([yi-fi]²)/N) = 2*sum(yi-fi)/N
        => dLoss/df ~ sum(yi - fi)/N
        """
        return predicted - target

    def fit(self, x: _pd.DataFrame, y: _np.ndarray, validation_data=None,
            validation_fraction=0.2, n_max_trees=100, shrinkage=0.1,
            patience=10, verbose=True, **kwargs):
        """
        Fit the model on the given data
        """
        # Getting some validation data
        if validation_data is None:
            data = _utils.train_test(x, y, validation_fraction)
            x_train, y_train, x_val, y_val = data
        else:
            x_train, y_train = x, y
            x_val, y_val = validation_data
        # initializing variables
        self.shrinkage = shrinkage
        self.trees = []
        target = y_train
        predicted_train = _np.zeros(y_train.shape)
        predicted_val = _np.zeros(y_val.shape)
        training_loss = []
        validation_loss = []
        best_loss = 0
        if verbose:
            print("Fitting gradient boosted regression tree:")
        # Training loop
        try:
            for n_trees in range(n_max_trees):
                tree = _RegressionTree()
                tree.fit(x_train, target, **kwargs)
                predicted_train += self.shrinkage*tree.predict(x_train)
                predicted_val += self.shrinkage*tree.predict(x_val)
                self.trees.append(tree)
                train_loss = self._MSE(predicted_train, y_train)
                training_loss.append(train_loss)
                val_loss = self._MSE(predicted_val, y_val)
                validation_loss.append(val_loss)
                target = -self._grad(predicted_train, y_train)
                if verbose:
                    print(f"\ttraining loss={train_loss:.3g}"
                          f"\tvalidation loss={val_loss:.3g}")
                if val_loss <= validation_loss[best_loss]:
                    best_loss = n_trees
                else:
                    if (n_trees - best_loss) > patience:
                        break
        except KeyboardInterrupt:
            pass
        # Removing excess DecisionTrees
        self.trees = self.trees[:best_loss+1]

    def predict(self, x: _pd.DataFrame):
        """
        predict the model
        """
        s = self.shrinkage
        return s*_np.sum([tree.predict(x) for tree in self.trees], axis=0)


class GradientBoostedClassificationTree(_ClassifierTemplate):
    """Gradient Boosted Classification Tree"""

    def __init__(self):
        self.shrinkage = 0.1
        self.trees = dict()

    def _softmax(self, output: _pd.DataFrame) -> _pd.DataFrame:
        """
        Calculate the softmax of the model's output.
        This converts the unborned model outputs for each categories
        to a probability 0 < pj < 1, such that sum_j(p_j) = 1.
            pij = exp(f[i,j]) / sum_j(exp(f[i,j]))
        with i the index of the observations,
        and j the index of the categories to predict,
        f[i,j] is the model's output at observation i for category j
        """
        exp = _np.exp(output.values)
        denominator = _np.sum(exp, axis=1)
        exp /= denominator[:, None]
        return _pd.DataFrame(data=exp, columns=self.categories)

    def _cross_entropy(cls, proba_predicted: _pd.DataFrame,
                       proba_target: _pd.DataFrame) -> float:
        """
        calculate the cross entropy between predicted/target probabilities.
            -sum_i,j(c[i,j] * ln(p[i,j]))
        which equals to:
            -sum_i(ln(p[i,k])
        where:
        i is the index over the observations,
        j is the index over the categories to predict,
        k is the target category of the ith observation,
        c[i,j] is the one-hot target vector (1 if j==k, 0 otherwise),
        p[i,j] is the predicted probability of category j for observation i
        """
        return -_np.mean(proba_target.values * _np.log(proba_predicted))

    def _log_likelihood(self, output: _pd.DataFrame,
                        target: _pd.DataFrame) -> float:
        """
        Returns the log-likelihood loss function from the non-normalized
        predictions and the target categories.
        It is calculated as the cross entropy of the softmaxed model's output.
        This reduces to:
            L = - sum_i,j( c[i,j] * ln( exp(f[i,j]) / sum_j(exp(f[i,j])) ) )
        which equals to:
            L = - sum_i(ln( exp(f[i,k]) / sum_j(exp(f[i,j]))))
        where:
        i is the index over the observations,
        j is the index over the categories to predict,
        k is the target category of the ith observation,
        c[i,j] is the one-hot target vector (1 if j==k, 0 otherwise),
        f[i,j] is the predicted value for observation i, class j
        (before the softmax is applied)
        """
        proba_predicted = self._softmax(output).values
        return self._cross_entropy(proba_predicted, target)

    def _grad(self, output: _pd.DataFrame,
              target: _pd.DataFrame) -> _pd.DataFrame:
        """
        returns the gradient (the Jacobian actually) of the loss function
        (the log-likelihood) relative to the model's prediction.
        It reduces to:
            dL/df[i,j] = p[i,j] - c[i,j]
        """
        proba_predicted = self._softmax(output)
        return proba_predicted - target

    def _to_categories(self, one_hot: _pd.DataFrame) -> _np.ndarray:
        """
        Converts a dataframe of one-hot encoded vectors to a list of
        categories.
        """
        return [self.categories[i] for i in _np.argmax(one_hot.values, axis=1)]

    def _to_one_hot(self, y: _np.ndarray) -> _pd.DataFrame:
        """
        Converts a vector of observed categories to a dataframe of one-hot
        encoded vectors.
        Each column of the dataframe is a category, each line corresponds to
        an observation.
        """
        y = _np.array(y)  # in case a _pd.Series is passed instead
        data = [[1. if c == y[j] else 0 for c in self.categories]
                for j in range(len(y))]
        return _pd.DataFrame(data=data, columns=self.categories)

    def _scalar_output(self, x: _pd.DataFrame, category: str):
        """
        Calculate the raw scalar output for a given category,
        for all observations x
        """
        s = self.shrinkage
        c = category
        return s*_np.sum([tree.predict(x) for tree in self.trees[c]], axis=0)

    def _output(self, x: _pd.DataFrame) -> _pd.DataFrame:
        """
        Calculate the raw output of the model without converting back
        to categories nor probabilities
        """
        vectorial = _np.transpose([self._scalar_output(x, c)
                                  for c in self.categories])
        return _pd.DataFrame(data=vectorial, columns=self.categories,
                             dtype=float)

    def fit(self, x: _pd.DataFrame, y: _np.ndarray, validation_data=None,
            validation_fraction=0.2, n_max_trees=100, shrinkage=0.1,
            patience=10, verbose=True, **kwargs):
        """
        Fit the model on the given data
        """
        # Getting some validation data
        if validation_data is None:
            data = _utils.train_test(x, y, validation_fraction)
            x_train, y_train, x_val, y_val = data
        else:
            x_train, y_train = x, y
            x_val, y_val = validation_data
        # initializing variables
        self.categories = _np.unique(_np.concatenate([y_train, y_val]))
        self.trees = {c: [] for c in self.categories}
        val_target = self._to_one_hot(y_val)
        train_target = self._to_one_hot(y_train)
        target = train_target
        training_loss = []
        validation_loss = []
        best_loss = 0
        if verbose:
            print("Fitting gradient boosted classification tree:")
        # Training loop
        try:
            for n_trees in range(n_max_trees):
                for c in self.categories:
                    tree = _RegressionTree()
                    tree.fit(x_train, target[c], **kwargs)
                    self.trees[c].append(tree)
                    if verbose:
                        print(".", end="", flush=True)
                output = self._output(x_train)
                train_loss = self._log_likelihood(output, train_target)
                training_loss.append(train_loss)
                val_loss = self._log_likelihood(self._output(x_val),
                                                val_target)
                validation_loss.append(val_loss)
                target = -self._grad(output, train_target)
                if verbose:
                    print(f"\ttraining loss={train_loss:.3g}"
                          f"\tvalidation loss={val_loss:.3g}")
                if val_loss <= validation_loss[best_loss]:
                    best_loss = n_trees
                else:
                    if (n_trees - best_loss) > patience:
                        break
        except KeyboardInterrupt:
            pass
        # Removing excess DecisionTrees
        for c in self.categories:
            self.trees[c] = self.trees[c][:best_loss+1]

    def predict(self, x: _pd.DataFrame) -> _np.ndarray:
        """
        Predict the model on the given data
        """
        return self._to_categories(self._output(x))



import machine_learning.neural_networks.common as _common
import machine_learning._templates as _templates
import torch as _torch
import types as _types
import machine_learning as _ml
import machine_learning.neural_networks.layers as _layers
import numpy as _np


class Classifier(_common.NNtemplate, _templates.ClassifierTemplate):
    """
    A model that predict scalar values from vectors of observations.

    Attributes:
    -----------
    x : list of str
        name of x observations columns
    categories : list
        categories the inputs can be classified into
    mean : np.ndarray
        The mean value for normalization of input
    std : np.ndarray
        The standard deviation for normalization of input
    dense_layers : list
        list of the dense layers of the model
    args : list or None
        list of arguments used to generate the model
    """

    def __init__(self):
        super().__init__()
        self.dense_layers = []
        self.loss = _torch.nn.functional.cross_entropy
        self.args = None

    def fit(self, observations, labels, layers=(32, 16, 8), validation=None,
            validation_fraction=0.2, restart=True, non_linear="relu",
            l_max=None, **kwargs):
        """
        Fit the model on the given data.

        Parameters:
        -----------
        observations : pd.DataFrame
            real valued observations
        labels : list
            label associated to each observation
        layers : list of int
            The number of neurons in each dense hidden layer
        validation : pandas.DataFrame or None
            data used for validation
        validation_fraction : float
            if validation is None, training data is splited
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function
        l_max : int or None
            maximum observations in a batch

        Returns:
        --------
        training, validation : pandas.DataFrame

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        training = observations, labels
        # preprocess
        training, validation = self._validation_split(training, validation,
                                                      validation_fraction)
        res = self._preprocess(training, validation)
        self.mean, self.std, self.categories, self.x, _, _ = res
        # create layers
        if restart:
            self._set_layers(layers, non_linear)
        # load data in memory
        train = self._to_tensor(training)
        val = self._to_tensor(validation)
        train, n_training = self._wrap_batch(train, l_max)
        val, n_validation = self._wrap_batch(val, l_max)
        # fitting
        self.train_loop(train, val, n_training, n_validation, **kwargs)
        # returning training and validation data
        return training, validation

    def fit_batch(self, training, validation, x, y, layers,
                  restart=True, non_linear="relu", **kwargs):
        """
        Fit the model on the given batches.

        Parameters:
        -----------
        training : generator function
            function that yields a pandas dataframe
        validation : generator function
            function that yields a pandas dataframe
        x : list of str
            a list of column name for x variables
        y : str
            column name of the prediction
        layers : list of int
            The number of neurons in each dense hidden layer
        restart : bool
            if False, the architecture is not modified
        non_linear : str or function
            non-linear function

        Other parameters:
        -----------------
        **kwargs from self.train_loop(...)
        """
        # preprocess
        res = self._preprocess(training, validation)
        self.mean, self.std, self.categories, self.x, \
            n_training, n_validation = res
        # create layers
        if restart:
            self._set_layers(layers, non_linear)
        # fitting
        self.train_loop(self._wrap(training), self._wrap(validation),
                        n_training, n_validation, **kwargs)

    def predict(self, observations):
        """
        Make a prediction using the trained model.

        Parameters:
        -----------
        observations : pandas.DataFrame

        Returns:
        --------
        labels : list
        """
        X = observations[self.x].values
        tensor = self._array_to_tensor(X, self.mean, self.std)
        with _torch.no_grad():
            output = self(tensor)
        labels = self._tensor_to_labels(output)
        return labels

    def forward(self, tensor):
        """Don't use this"""
        for i, dense in enumerate(self.dense_layers):
            tensor = dense(tensor)
        return tensor

    def _to_tensor(self, data):
        """Converts a dataframe into a tuple of tensors"""
        X, Y = data
        X = X[self.x].values
        X = self._array_to_tensor(X, mean=self.mean, std=self.std)
        Y = self._labels_to_tensor(Y)
        return X, Y

    def _preprocess(self, training, validation):
        """preprocess the data"""
        # batchify if needed
        if not isinstance(training, _types.FunctionType):
            return self._preprocess(lambda: (yield training),
                                    lambda: (yield validation))
        # treat by batches
        sum_, sum2, n = 0., 0., 0.
        categories = set()
        x = None
        # treat training data
        for i, (obs, labels) in enumerate(training()):
            if x is None:
                x = list(obs.columns)
            for key in obs.columns:
                if obs[key].isna().any():
                    raise ValueError(f"At batch {i}: Found NaN in "
                                     f"training x['{key}']")
            sum_, sum2, n = self._sums(obs.values, sum_, sum2, n)
            categories = categories | set(labels)
        n_training = i+1
        # treat validation data
        for i, (obs, labels) in enumerate(validation()):
            for key in obs.columns:
                if obs[key].isna().any():
                    raise ValueError(f"At batch {i}: Found NaN in "
                                     f"validation x['{key}']")
            sum_, sum2, n = self._sums(obs.values, sum_, sum2, n)
            categories = categories | set(labels)
        n_validation = i+1
        # returns values
        mean, std = self._mean_std(sum_, sum2, n)
        return mean, std, list(categories), x, n_training, n_validation

    def _set_layers(self, *args):
        """Remove the previous layers and create new ones"""
        layers, non_linear = args
        self.args = args
        # Deleting previous layers
        for i, _ in enumerate(self.dense_layers):
            delattr(self, f"dense{i}")
        self.dense_layers = []
        # Set dense layers
        dens = _layers.DenseLayer
        n = len(self.x)
        for i, n_out in enumerate(layers):
            layer = dens(n, n_out, non_linear=non_linear)
            self.dense_layers.append(layer)
            setattr(self, f"dense{i}", layer)
            n = layer.shape_out
        # out layer
        i = len(self.dense_layers)
        n_out = len(self.categories)
        layer = dens(n, n_out, non_linear="identity")
        self.dense_layers.append(layer)
        setattr(self, f"dense{i}", layer)
        # migrate to appropriate device
        self.to(self.device)
        # reset optimization
        self.reset()

    @property
    def dump(self):
        parameters = dict()
        parameters["args"] = self.args
        parameters["x"] = self.x
        parameters["categories"] = self.categories
        parameters["mean"] = self.mean.tolist()
        parameters["std"] = self.std.tolist()
        dense = [d.dump for d in self.dense_layers]
        parameters["dense_layers"] = dense
        return {type(self).__name__: parameters, "version": _ml.__version__}

    @dump.setter
    def dump(self, other):
        if type(self).__name__ not in other.keys():
            raise ValueError(f"Expected dump from a '{type(self).__name__}'")
        other = other[type(self).__name__]
        self.x = other["x"]
        self.categories = other["categories"]
        self.mean = _np.array(other["mean"])
        self.std = _np.array(other["std"])
        self.args = other["args"]
        if self.args is not None:
            self._set_layers(*self.args)
        for i, d in enumerate(other["dense_layers"]):
            self.dense_layers[i].dump = d
        self.reset_optimizer()



import numpy as _np
import pandas as _pd
from machine_learning._templates import RegressorTemplate as _Regressor
from machine_learning._templates import ClassifierTemplate as _Classifier
import machine_learning.library as _lib
from machine_learning.library import decision_trees as _dt


class Branch:
    """ A class representing a branch of a dichotomous tree

    Attributes:
    -----------
    child_true : branch or None
        The child that verifies the condition.
        If None, this branch is a leaf.
    child_false : branch or None
        The child that doesn't verify the condition.
        If None, this branch is a leaf.
    variable : str
        The Name of the variable on which the condition applies.
        This is the left hand sign of the test.
    criterion : str
        The criterion of the test ('<=' or '==')
    threshold : float
        The threshold value for the right hand sign of the test.
    gain : float
        The gain of the split of this branch
    value : object
        For leafs, the value predicted for items that reach the leaf
    mask : np.ndarray
        An array of boolean masking the data that reached this branch
    """

    operation = {"<=": lambda x, y: x <= y,
                 "==": lambda x, y: x == y}

    def __init__(self):
        self.child_true = None
        self.child_false = None
        self.variable = "name"
        self.criterion = "<="
        self.threshold = float("nan")
        self.gain = 0.
        self.value = None
        self.mask = None

    def from_cBranch(self, cbranch: _dt.cBranch, x_num: _pd.DataFrame,
                     x_cat: _pd.DataFrame, y: _np.ndarray):
        """fill branch from the data from a cBranch"""
        if cbranch.is_leaf:
            if (y.dtype == float):
                self.value = cbranch.y_num
            else:
                self.value = y[cbranch.iy_cat]
        else:
            x = x_num if cbranch.split_numerical_x else x_cat
            self.variable = x.columns[cbranch.j]
            self.criterion = "<=" if cbranch.split_numerical_x else "=="
            self.threshold = x[self.variable].iloc[cbranch.i]
            self.gain = cbranch.gain

    def propagate(self, x: _pd.DataFrame, mask: _np.ndarray):
        """Recursively propagate the dataframe in the tree."""
        self.mask = mask
        if not self.is_leaf:
            operator = self.operation[self.criterion]
            test = operator(x[self.variable], self.threshold)
            self.child_true.propagate(x, mask & test)
            self.child_false.propagate(x, mask & ~test)

    def grow(self, ctree: list, index: int, leafs: list,
             x_num: _pd.DataFrame, x_cat: _pd.DataFrame, y: _np.ndarray):
        """Recursively expands the tree to copy the content of the cTree."""
        cbranch = ctree[index]
        self.from_cBranch(cbranch, x_num, x_cat, y)
        if not cbranch.is_leaf:
            self.child_true = Branch()
            self.child_true.grow(ctree, cbranch.child_true, leafs, x_num,
                                 x_cat, y)
            self.child_false = Branch()
            self.child_false.grow(ctree, cbranch.child_false, leafs, x_num,
                                  x_cat, y)
        else:
            leafs.append(self)

    @property
    def is_leaf(self):
        """ True if this branch is a leaf"""
        return (self.child_true is None) and (self.child_false is None)


class DecisionTree:

    def split_num_cat(x: _pd.DataFrame):
        """
        split x into two dataframes:
            - one of numerical data
            - one of categorical data
        """
        is_numerical = [_np.issubdtype(x[c].dtype, _np.number)
                        for c in x.columns]
        c_num = [c for i, c in enumerate(x.columns) if is_numerical[i]]
        c_cat = [c for i, c in enumerate(x.columns) if not is_numerical[i]]
        x_num = x[c_num]
        x_cat = x[c_cat].fillna("")
        return x_num, x_cat

    def __init__(self):
        super().__init__()
        self.root = Branch()
        self.leafs = []

    def grow(self, ctree: list, x_num: _pd.DataFrame, x_cat: _pd.DataFrame,
             y: _np.ndarray):
        """Expand the root of a tree"""
        self.root.grow(ctree, 0, self.leafs, x_num, x_cat, y)

    def predict(self, x: _pd.DataFrame):
        "predict y for the input x"
        if hasattr(self, "categories"):
            L = _np.max([len(s) for s in self.categories])
            y_pred = _np.ndarray(len(x), dtype=f"U{L}")
        else:
            y_pred = _np.ndarray(len(x), dtype=float)
        self.root.propagate(x, _np.array([True]*len(x)))
        for leaf in self.leafs:
            y_pred[leaf.mask] = leaf.value
        return y_pred


class ClassificationTree(DecisionTree, _Classifier):

    def __init__(self):
        super().__init__()

    def fit(self, x: _pd.DataFrame, y: _np.ndarray, **kwargs):
        if isinstance(y, _pd.Series):
            y = y.to_numpy()
        y = y.astype(str)
        self.categories = _np.unique(y)
        x_num, x_cat = DecisionTree.split_num_cat(x)
        ctree = _dt.grow_classifier_tree(x_num.to_numpy(float),
                                         x_cat.to_numpy(str),
                                         y, **kwargs)
        branches = [ctree.branches[i] for i in range(ctree.n_branches)]
        self.grow(branches, x_num, x_cat, y)
        _lib.delete_tree(ctree)


class RegressionTree(DecisionTree, _Regressor):

    def __init__(self):
        super().__init__()

    def fit(self, x: _pd.DataFrame, y: _np.ndarray, **kwargs):
        if isinstance(y, _pd.Series):
            y = y.to_numpy()
        y = y.astype(float)
        assert not _np.isnan(y).any()
        x_num, x_cat = DecisionTree.split_num_cat(x)
        y_pred = _np.ndarray(len(y), float)  # debug
        ctree = _dt.grow_regressor_tree(x_num.to_numpy(float).copy(order='C'),
                                        x_cat.to_numpy(str).copy(order='C'),
                                        y, y_pred=y_pred, **kwargs)
        branches = [ctree.branches[i] for i in range(ctree.n_branches)]
        self.grow(branches, x_num, x_cat, y)
        _dt.delete_tree(ctree)



import torch
import pandas as pd
import numpy as np
import torch.nn.functional as F
from typing import List, Union, Iterable
from .layers import FullyConnected, BatchNorm1d, Linear
from .conversions import dataframe_to_tensor, \
                         floats_to_tensor, tensor_to_floats
from .nn_decorators import neural_network


@neural_network
class DenseRegressor(torch.nn.Module):

    @classmethod
    def from_dump(cls, dump):
        pass

    def __init__(self, inputs: List[str],
                 hidden_layers: List[int] = [10, 10, 10],
                 activation: str = "relu"):
        super().__init__()
        self.inputs = list(inputs)
        self.input_norm = BatchNorm1d(len(inputs))
        self.fully_connected = FullyConnected(len(inputs),
                                              hidden_layers=hidden_layers,
                                              activation=activation)
        self.output = Linear(self.fully_connected.out_features, 1)
        self.target_norm = BatchNorm1d(1)

    def forward(self, x):
        x = self.input_norm(x)
        x = self.fully_connected(x)
        return self.output(x)

    def data_to_tensor(self, X: Union[pd.DataFrame, Iterable],
                       Y: Union[None, np.ndarray],
                       weights: Union[None, List[float]] = None
                       ) -> tuple:
        if isinstance(X, pd.DataFrame):
            x = dataframe_to_tensor(X, self.inputs, self.device)
        else:
            x = floats_to_tensor(X, self.device)
        y = None if Y is None else floats_to_tensor(Y, self.device).view(-1, 1)
        w = None if weights is None else floats_to_tensor(weights).view(-1, 1)
        return x, y, w

    def tensor_to_y(self, tensor: torch.Tensor) -> np.ndarray:
        return tensor_to_floats(self.target_norm.undo(tensor).view(-1))

    def loss(self, y_pred: torch.Tensor, y_target: torch.Tensor,
             weights: Union[None, torch.Tensor]) -> torch.Tensor:
        y_target = self.target_norm(y_target)
        if weights is None:
            return torch.sqrt(F.mse_loss(y_pred, y_target))
        else:
            return torch.sqrt(torch.mean(weights * (y_pred - y_target)**2))

    @property
    def dump(self):
        pass
